//==============================================================================
//  CRYSTALFONTZ 
//==============================================================================
//
//  This code drives the touch screen on the CFAF480128A0-039TC display
//    https://www.crystalfontz.com/product/cfaf480128a0039tc
//
//  Seeeduino v4.2, an open-source 3.3v capable Arduino clone.
//    https://www.seeedstudio.com/Seeeduino-V4.2-p-2517.html
//    https://github.com/SeeedDocument/SeeeduinoV4/raw/master/resources/Seeeduino_v4.2_sch.pdf
//
//==============================================================================
//
//  2019-02-14 Trevin Jorgenson
//
//==============================================================================
//This is free and unencumbered software released into the public domain.
//
//Anyone is free to copy, modify, publish, use, compile, sell, or
//distribute this software, either in source code form or as a compiled
//binary, for any purpose, commercial or non-commercial, and by any
//means.
//
//In jurisdictions that recognize copyright laws, the author or authors
//of this software dedicate any and all copyright interest in the
//software to the public domain. We make this dedication for the benefit
//of the public at large and to the detriment of our heirs and
//successors. We intend this dedication to be an overt act of
//relinquishment in perpetuity of all present and future rights to this
//software under copyright law.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
//OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
//ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
//OTHER DEALINGS IN THE SOFTWARE.
//
//For more information, please refer to <http://unlicense.org/>
//==============================================================================

#include <Wire.h>
#include <stdarg.h>
#include "tma568.h"

//We need to receive up to 110 bytes and so the I2C buffer needs to be modified:
//C:\Program Files(x86)\Arduino\hardware\arduino\avr\libraries\Wire\src\Wire.h
//line 28 has   #define BUFFER_LENGTH 32 which needs to be changed to 128
//
//C:\Program Files(x86)\Arduino\hardware\arduino\avr\libraries\Wire\src\utilit\twi.h
//line 32 has   #define TWI_BUFFER_LENGTH 32 which needs to be changed to 128
//


#define SET_DBG     (PORTB |=  0x04)
#define CLR_DBG     (PORTB &= ~0x04)

#define TCH_RES     8
#define INTERRUPT   9
#define I2C_SDA     18
#define I2C_SCL     19

#define OPTIMIZED_COMMANDS

#ifdef OPTIMIZED_COMMANDS

#define SET_RESET   (PORTB |=  0x01)
#define CLR_RESET   (PORTB &= ~0x01)
#define SET_INT     (PORTB |=  0x02)
#define CLR_INT	    (PORTB &= ~0x02)
#else
#define SET_RESET   digitalWrite(TCH_RES,HIGH)
#define CLR_RESET   digitalWrite(TCH_RES,LOW)
#define SET_INT     digitalWrite(INTERRUPT,HIGH)
#define CLR_INT	    digitalWrite(INTERRUPT,LOW)
#endif


//==============================================================================
// Cap Touch control lines
//   ARD      | Port  | CFAF480128A0-TS |  Function - Parallel        
//------------+-------+-----------------+-----------------------------
//  3.3V      |       | #2              |  POWER 3.3V                 
//  GND	      |       | #1,10           |  GROUND                     
// -----------+-------+-----------------+-----------------------------
//  D8	      | PORTC | #7              |  Reset          (TCH_RESET) 
//  D9        | PORTC | #9              |  Interrupt      (INTERRUPT)
// -----------+-------+-----------------+-----------------------------
//  A4/SDA    | PORTC | #5              |  I2C SDA           (I2C_SDA)
//  A5/SCL    | PORTD | #3              |  I2C SCL           (I2C_SCL)              
// -----------+-------+-----------------+-----------------------------
//==============================================================================
//  Note: For I2C protocol to work, pullup resistors need to be attached to SDA and SCL
//        refer to I2C protocol
//  Note: Due to a lack of pins on the Arduino, the Reset pin was set to 3.3v but
//        this should be tied to an available GPIO
//==============================================================================

//==============================================================================
// ref http://playground.arduino.cc/Main/Printf
//
// Example to dump a uint32_t in hex and decimal
// SerPrintFF(F("RAM_G_Unused_Start: 0x%08lX = %lu\n"),RAM_G_Unused_Start,RAM_G_Unused_Start);
// Example to dump a uint16_t in hex and decimal
// SerPrintFF(F("Initial Offest Read: 0x%04X = %u\n"),FT8xx_write_offset ,FT8xx_write_offset);
void SerPrintFF(const __FlashStringHelper *fmt, ...)
{
  char
    tmp[128]; // resulting string limited to 128 chars
  va_list
    args;
  va_start(args, fmt);
  vsnprintf_P(tmp, 128, (const char *)fmt, args);
  va_end(args);
  Serial.print(tmp);
}


//==============================================================================
// readTouch() reads the touch data from the touch controller via I2C
// and stores the data in the struct passed into it
void readTouch(TMA568RETURNSTRUCT *touch)
{
  uint8_t temp;
  //Start trasmission by sending address 0x24
  Wire.beginTransmission(0x24);
  //Start reading at register 0x0003
  Wire.write(0x03);
  Wire.write(0x00);
  //Send the data over
  Wire.endTransmission();

  Wire.requestFrom((uint8_t)0x24, (uint8_t)sizeof(TMA568RETURNSTRUCT));

  //Read in byte 0, 1 = Number of valid bytes in packet
  uint16_t recordLength;
  recordLength = Wire.read();
  recordLength |= (Wire.read() << 8);
  //if the record length is less than 7 we know there are no touches
  if (recordLength < 7)
  {
    touch->numpoints = 0;
    return;
  }

  //Dispose of byte 2 = record ID
  Wire.read();

  //Dispose of byte 3, 4 = Timestamp
  Wire.read();
  Wire.read();

  //Read in byte 5
  temp = Wire.read();
  //Extract the large Object flag
  touch->largeobject = (0 != (temp & 0x20)) ? 1 : 0;

  //Extract Number of Records
  touch->numpoints = temp & 0x1f;
  //limit this to 9 points so we don't overflow our struct
  if (9 < touch->numpoints)
  {
    touch->numpoints = 9;
  }

  //Dispose of byte 6 = report counter & noise effects
  Wire.read();

  //Loop through the nine touch records
  for (uint8_t i = 0; i < touch->numpoints; i++)
  {
    //Read in byte 0, Extract Number of Records
    touch->points[i].touchtype = Wire.read() & 0x07;

    //Read in byte 1, 
    temp = Wire.read();
    //Extract Tip = Lift off status
    touch->points[i].tip = (0 != (temp & 0x80)) ? 1 : 0;;
    //Extract Event ID
    touch->points[i].eventid = (temp & 0x60) >> 5;
    //Extract Touch ID
    touch->points[i].touchid = temp & 0x1f;

    //Read in byte 2, LSB of X
    touch->points[i].X = Wire.read();
    //Read in byte 3, MSB of X
    touch->points[i].X |= (Wire.read() << 8);
    //uint16_t bigTemp;
    //bigTemp = Wire.read();
    //bigTemp |= (Wire.read() << 8);
    //Serial.println(bigTemp);

    //Read in byte 4, LSB of Y
    touch->points[i].Y = Wire.read();
    //Read in byte 5, MSB of Y
    touch->points[i].Y |= (Wire.read() << 8);
    //Serial.println(touch->points[i].Y);

    //Read in byte 6, Pressure
    touch->points[i].pressure = Wire.read();

    //Read in byte 7, Major Axis Length
    touch->points[i].major = Wire.read();

    //Read in byte 8, Minor Axis Length
    touch->points[i].minor = Wire.read();

    //Read in byte 9, Orientation
    touch->points[i].orientation = Wire.read();
  }

}

//==============================================================================
// pre-loop setup
void setup()
{
  DDRB = 0x05;
  //SET_INT;
  Serial.begin(115200);
  Serial.println("serial initialized");

  Wire.begin();
  Wire.setClock(400000);


  digitalWrite(TCH_RES, LOW);
  delay(500);
  digitalWrite(TCH_RES, HIGH);
  delay(200);
}


//==============================================================================
// main loop
void loop()
{
  //Create an object to hold all of the touch data
  TMA568RETURNSTRUCT
    thisTouchReading;

  //Wait for interrupt pin to signal a packet is ready
  while (digitalRead(INTERRUPT) == HIGH);

  //Specs on the transaction packets
  //Total transaction time is ~3.25ms with 110 byte buffer size @ 
  //Interrupt goes low for 1.5ms
  //Maximum Report rate ~7.4ms  

  readTouch(&thisTouchReading);
  SerPrintFF(F("NumPoints: %d"), thisTouchReading.numpoints);
  for (uint8_t i = 0; i < thisTouchReading.numpoints; i++)
  {
    SerPrintFF(F(" | Pt[%d,%d]=(%3d,%3d) "), i, thisTouchReading.points[i].eventid, thisTouchReading.points[i].X, thisTouchReading.points[i].Y);
  }
  Serial.println();
}